// Copyright (C) 2000 - 2002 Hewlett-Packard Company
//
// This program is free software; you can redistribute it and/or modify it
// under the term of the GNU Lesser General Public License as published by the
// Free Software Foundation; either version 2 of the License, or (at your
// option) any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
// for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program; if not, write to the Free Software Foundation,
// Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
// _________________

// @(#) $Revision: 4.5 $ $Source: /judy/src/apps/demo/funhist.c $
//
// FUNCTION HISTOGRAM; JUDYL DEMO PROGRAM
//
// This program uses JudyL to create a histogram of data generated by the
// function random(3M).  Other functions could be substituted in.
//
// The program generates n random numbers, stores them in a JudyL array
// and then prints a histogram.  All aspects of the generation and
// histogram are timed so the user can see how long various JudyL functions
// take.
//
//
// This demonstrates:
//           JudyLIns
//           JudyLFirst
//           JudyLNext
//           JudyLLast
//           JudyLCount
//           JudyLByCount
//           how to access a JudyL value.
//
// Notice that using JudyL gives you fast stores and lookups as with hashing
// but without having to define a hash function, initialize the hash table
// or having to predetermine a good hash table size.  It also gives you a
// sorted list at the same time.
// Also notice that JudyLCount is much faster than you can sequentially
// count items in an array.
//

#include <stdio.h>
#include <sys/time.h>
#include <errno.h>

#include "Judy.h"

// Default number of iterations (number of random numbers generated)
// This may be overridden on the command line

#define DEFAULT_ITER 1000000

// The number of buckets the histogram is divided into.

#define DEFAULT_HISTO_BUCKETS 32

// Macro for correction english output for plurals

#define	PLURAL(count) ((count == 1) ? "" : "s")

// Generic fatal error handler, pass in JError_t

#define JUDY_ERROR(J) \
	fprintf(stderr, "File: '%s', line: %d: Judy error: %d\n", \
                __FILE__, __LINE__, JU_ERRNO(&(J))); \
        exit(2);

// timing routines

struct timeval TBeg, TEnd;
#define STARTTm         gettimeofday(&TBeg, NULL)
#define ENDTm           gettimeofday(&TEnd, NULL)
#define DeltaUSec \
        ((double)(TEnd.tv_sec * 1000000.0 + TEnd.tv_usec) \
        - (TBeg.tv_sec * 1000000.0 + TBeg.tv_usec))
// End of timing routines


main( int argc, char **argv)
{
   void *   PJArray = 0;	// main JudyL array
				// key is random number, value is repeat count
   void *   PJCountArray = 0;	// second JudyL array
				// key is repeat count (from PJArray)
				// value is count of items with the same
				// repeat count
   JError_t JError;             // Judy error structure
   Word_t   Index;		// index in JudyLFirst/Next loop
   PWord_t  PValue;		// pointer to Judy array value
   PWord_t  PGenCount;		// pointer to generation count
   Word_t   num_vals;		// number of randoms to generate
   Word_t   iter;		// for loop iteration
   Word_t   unique_nums;        // number of unique randoms generated
   Word_t   random_num;         // random number
   Word_t   median;             // random number
   Word_t   tmp1, tmp2;         // temporary variable
   double   random_gen_time;	// time to generate random numbers
   Word_t   histo_incr;		// histogram increment
   unsigned long long ran_sum = 0;  // sum of all generated randoms

// TITLE

   (void) puts ("\nJudyL demonstration: random(3M) histogram");

// SET NUMBER OF RANDOMS TO GENERATE

   if (argc != 2)
   {
      // SET TO DEFAULT_ITER
      num_vals = DEFAULT_ITER;
      (void) printf ("Usage: %s [numvals]\n", argv[0]);
      (void) printf ("  Since you did not specify a number of values, %d\n",
                     num_vals);
      (void) puts   ("  will be used as the number of random numbers to insert");
      (void) puts   ("  into the Judy array");
   }
   else
   {
// OVERRIDE NUMBER OF RANDOMS TO GENERATE

      num_vals = atol(argv[1]);
   }

// TIME THE GENERATION OF ALL THE RANDOM NUMBERS.  THIS TIME IS LATER
//
// This time is later subtracted from the insert loop time so that the
// time it takes to do the actual JudyLIns can be isolated from the
// total time.

   (void) puts ("\ntiming random number generation");
   STARTTm;
   for (iter = num_vals; iter; iter--)
   {
      random();
   }  /* end of random number generator time */
   ENDTm;
   random_gen_time = DeltaUSec;
   (void) printf("It took %.3f sec to generate %ld random numbers\n",
     random_gen_time/1000000, num_vals);
   (void) printf("  (ie. %.3f uSec/number)\n\n", random_gen_time/num_vals);


// REGENERATE RANDOMS AND INSERT THEM INTO A JUDYL ARRAY

   (void) puts("Please wait while the random numbers are inserted into");
   (void) puts("a JudyL array (with a usage count) ...");

   STARTTm;
   for (iter = num_vals; iter; iter--)
   {
      random_num = (Word_t) random();

      if ((PValue = (PWord_t) JudyLIns(&PJArray, random_num, &JError))
       == (PWord_t) PJERR)
      {
         JUDY_ERROR(JError);  // exits
      }

      /* increment hit count */
      (*PValue)++;

      /* sum the random number */
      ran_sum += random_num;
   }  /* end of random number generator time */
   ENDTm;
   (void) printf("That took %.3f uSec/Index.\n",
		 (DeltaUSec - random_gen_time)/num_vals);


// COUNT THE NUMBER OF ELEMENTS IN THE JUDYL ARRAY
// IE. COUNT THE NUMBER OF UNIQUE RANDOM NUMBERS

   STARTTm;
   unique_nums = JudyLCount(PJArray, 0, ~0L, &JError);
   ENDTm;
   (void) printf("\nThere were %ld unique random numbers generated\n",
		unique_nums);
   (void) printf("It took %.3f uSec to count them in the Judy array.\n",
		DeltaUSec);


// FIND HOW MANY NUMBERS WERE GENERATED ONCE, TWICE, ...
//
// Create a new JudyL array where the index is the count from PJArray
// and the value is a count of the number of elements with that count.

   if (unique_nums != num_vals)
   {
      (void) puts("\nLooping through the entire Judy array to create a");
      (void) puts("new Judy counting array (PJCountArray in the source code)");
      (void) puts("...");
      STARTTm;
      for (Index = 0L,
           PValue  = (PWord_t) JudyLFirst(PJArray, &Index, &JError);
           PValue != (PWord_t) NULL;
           PValue  = (PWord_t) JudyLNext( PJArray, &Index, &JError))
      {
          if ((PGenCount = (PWord_t) JudyLIns(&PJCountArray, *PValue, &JError))
	   == (PWord_t) PJERR)
          {
             JUDY_ERROR(JError);  // exits
          }
          (*PGenCount)++; // increment hit count
      }
      ENDTm;

      (void) printf("That took %.3f Secs or %.3f uSec/Index\n\n",
			DeltaUSec/1000000, DeltaUSec/unique_nums);

// PRINT DUPLICATES HISTOGRAM

      (void) puts("Duplicates Histogram:");
      for (Index = 0L,
           PValue  = (PWord_t) JudyLFirst(PJCountArray, &Index, &JError);
           PValue != (PWord_t) NULL;
           PValue  = (PWord_t) JudyLNext( PJCountArray, &Index, &JError))
      {
          (void) printf("  %ld numbers were generated %ld time%s\n",
			*PValue, Index, PLURAL(Index));
      }
   }


// PRINT DISTRIBUTION HISTOGRAM

   (void) puts("\nCompute the random number distribution by counting index");
   (void) puts("ranges.");

   histo_incr = ((Word_t) ~0L / DEFAULT_HISTO_BUCKETS) >> 1;

   Index = 0L;
   for (iter = 0; iter < DEFAULT_HISTO_BUCKETS; iter++)
   {
      (void) printf("  %ld unique values from %8x - %8x\n",
		    JudyLCount(PJArray, Index, Index + histo_incr, &JError),
		    Index, Index + histo_incr);
      Index += histo_incr + 1;

   }


// PRINT MEAN (average),
//       MEDIAN (middle value, or average of two middle values)
//       RANGE (low and high value)

   tmp1 = (Word_t) (ran_sum/(long long)num_vals);
   (void) printf("                  mean: 0x%08x\n", tmp1);

// If there were an even number of randoms generated, then average
// the two middle numbers.  Otherwise, the mean is the middle value
   if (num_vals & 1)
   {
      JudyLByCount(PJArray, num_vals/2, &tmp1, &JError);
      JudyLByCount(PJArray, (num_vals/2)+1, &tmp2, &JError);
      median = (tmp1 + tmp2) / 2;
   }
   else
   {
      JudyLByCount(PJArray, (num_vals+1)/2, &median, &JError);
   }
   (void) printf("                median: 0x%08x\n", median);

   Index = 0;
   JudyLFirst(PJArray, &Index, 0);
   (void) printf("first random generated: 0x%08x\n", Index);

   Index = ~0;
   JudyLLast(PJArray, &Index, 0);
   (void) printf(" last random generated: 0x%08x\n", Index);

   exit(0);
   /*NOTREACHED*/

}  // main()
